/*
	*
	* Copyright 2019 Britanicus <marcusbritanicus@gmail.com>
	*
	* This file is a part of DesQ project (https://gitlab.com/desq/)
	*

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, write to the Free Software
    * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    * MA 02110-1301, USA.
    *
*/

#include "DesQUtils.hpp"
#include "desq-config.h"

static QMimeDatabase mimeDb;

namespace Size {
	const double KiB = 1024.0;
	const double MiB = 1024.0 * KiB;
	const double GiB = 1024.0 * MiB;
	const double TiB = 1024.0 * GiB;
};

void Logger( QtMsgType type, const QMessageLogContext &context, const QString &message ) {

	Q_UNUSED( context );

	QDir logDir( QDir::home().filePath( ".config/DesQ/" ) );
	if ( not logDir.exists() )
		DesQUtils::mkpath( logDir.path() );

	QString logPath = QDir::home().filePath( ".config/DesQ/DesQSession.log" );
	FILE *nblog = fopen( logPath.toLocal8Bit().data(), "a" );

	switch ( type ) {
		case QtInfoMsg: {
			fprintf( nblog, "DesQSession::Debug# %s\n", message.toLocal8Bit().data() );
			fflush( nblog );
			fprintf( stderr, "\033[01;32mDesQSession::Info# %s\n\033[00;00m", message.toLocal8Bit().data() );
			break;
		}

		case QtDebugMsg: {
			fprintf( nblog, "DesQSession::Debug# %s\n", message.toLocal8Bit().data() );
			fflush( nblog );
			fprintf( stderr, "\033[01;30mDesQSession::Debug# %s\n\033[00;00m", message.toLocal8Bit().data() );
			break;
		}

		case QtWarningMsg: {
			fprintf( nblog, "DesQSession::Warning# %s\n", message.toLocal8Bit().data() );
			fflush( nblog );
			if ( QString( message ).contains( "X Error" ) or QString( message ).contains( "libpng warning" ) )
				break;
			fprintf( stderr, "\033[01;33mDesQSession::Warning# %s\n\033[00;00m", message.toLocal8Bit().data() );
			break;
		}

		case QtCriticalMsg: {
			fprintf( nblog, "DesQSession::CriticalError# %s\n", message.toLocal8Bit().data() );
			fflush( nblog );
			fprintf( stderr, "\033[01;31mDesQSession::CriticalError# %s\n\033[00;00m", message.toLocal8Bit().data() );
			break;
		}

		case QtFatalMsg: {
			fprintf( nblog, "DesQSession::FatalError# %s\n", message.toLocal8Bit().data() );
			fflush( nblog );
			fprintf( stderr, "\033[01;41mDesQSession::FatalError# %s\n\033[00;00m", message.toLocal8Bit().data() );
			abort();
		}
	}
};

QString DesQUtils::dirName( QString path ) {

	if ( path == "/" or path == "//" )
		return "/";

	/* Simple clean path: remove '//' and './' */
	path = path.replace( "//", "/" ).replace( "/./", "/" );

	char *dupPath = strdup( path.toLocal8Bit().constData() );
	QString dirPth = QString::fromLocal8Bit( dirname( dupPath ) );
	dirPth += ( dirPth.endsWith( "/" ) ? "" : "/" );
	free( dupPath );

	return dirPth;
};

QString DesQUtils::baseName( QString path ) {

	if ( path == "/" or path == "//" )
		return "/";

	/* Simple clean path" remove '//' and './' */
	path = path.replace( "//", "/" ).replace( "/./", "/" );

	char *dupPath = strdup( path.toLocal8Bit().constData() );
	QString basePth = QString::fromLocal8Bit( basename( dupPath ) );
	free( dupPath );

	return basePth;
};

QString DesQUtils::getMimeType( QString path ) {

	return mimeDb.mimeTypeForFile( path ).name();
};

QString DesQUtils::getMimeIcon( QString path ) {

	return mimeDb.mimeTypeForFile( path ).iconName();
};

bool DesQUtils::isFile( QString path ) {

	struct stat statbuf;
	if ( stat( path.toLocal8Bit().constData(), &statbuf ) == 0 )

		if ( S_ISREG( statbuf.st_mode ) or S_ISLNK( statbuf.st_mode ) )
			return true;

		else
			return false;

	else
		return false;
};

bool DesQUtils::isDir( QString path ) {

	struct stat statbuf;
	if ( stat( path.toLocal8Bit().constData(), &statbuf ) == 0 )

		if ( S_ISDIR( statbuf.st_mode ) )
			return true;

		else
			return false;

	else
		return false;
};

bool DesQUtils::isLink( QString path ) {

	struct stat statbuf;
	if ( lstat( path.toLocal8Bit().constData(), &statbuf ) == 0 )
		if ( S_ISLNK( statbuf.st_mode ) )
			return true;

		else
			return false;

	else
		return false;
};

bool DesQUtils::exists( QString path ) {

	return not access( path.toLocal8Bit().constData(), F_OK );
}

QString DesQUtils::readLink( QString path ) {

	char linkTarget[ PATH_MAX ] = { 0 };
	realpath( path.toLocal8Bit().constData(), linkTarget );

	return QString( linkTarget );
};

int DesQUtils::mkpath( QString path, mode_t mode ) {

	/* Root always exists */
	if ( path == "/" )
		return 0;

	/* If the directory exists, thats okay for us */
	if ( DesQUtils::exists( path ) )
		return 0;

	mkpath( DesQUtils::dirName( path ), mode );

	return mkdir( path.toLocal8Bit().constData(), mode );
};

bool DesQUtils::removeDir( QString dirName ) {

	bool result = true;
	QDir dir( dirName);

	if ( dir.exists( dirName ) ) {
		Q_FOREACH( QFileInfo info, dir.entryInfoList( QDir::NoDotAndDotDot | QDir::System | QDir::Hidden | QDir::AllDirs | QDir::Files, QDir::DirsFirst ) ) {
			if ( info.isDir() )
				result = DesQUtils::removeDir( info.absoluteFilePath() );

			else
				result = QFile::remove( info.absoluteFilePath() );

			if ( !result )
				return result;
		}
		result = dir.rmdir( dirName );
	}

	return result;
};

bool DesQUtils::isReadable( QString path ) {

	if ( isDir( path ) )
		return not access( path.toLocal8Bit().constData(), R_OK | X_OK );

	else
		return not access( path.toLocal8Bit().constData(), R_OK );
};

bool DesQUtils::isWritable( QString path ) {

	return not access( path.toLocal8Bit().constData(), W_OK );
};

bool DesQUtils::isExecutable( QString path ) {

	struct stat statbuf;
	if ( stat( path.toLocal8Bit().data(), &statbuf ) != 0 )
		return false;

	if ( ( statbuf.st_mode & S_IXUSR ) ) {
		QMimeType m = mimeDb.mimeTypeForFile( path );
		if ( m.name() == "application/x-executable" )
			return true;

		else if ( m.name() == "application/x-sharedlib" )
			return true;

		else if ( m.allAncestors().contains( "application/x-executable" ) )
			return true;

		/* Default is false */
		return false;
	}

	return false;
};

qint64 DesQUtils::getSize( QString path ) {

	struct stat statbuf;
	if ( stat( path.toLocal8Bit().constData(), &statbuf ) != 0 )
		return 0;

	switch( statbuf.st_mode & S_IFMT ) {
		case S_IFREG: {

			return statbuf.st_size;
		}

		case S_IFDIR: {
			DIR* d_fh;
			struct dirent* entry;
			QString longest_name;

			while ( ( d_fh = opendir( path.toLocal8Bit().constData() ) ) == NULL ) {
				qWarning() << "Couldn't open directory:" << path;
				return statbuf.st_size;
			}

			quint64 size = statbuf.st_size;

			longest_name = QString( path );
			if ( not longest_name.endsWith( "/" ) )
				longest_name += "/";

			while( ( entry = readdir( d_fh ) ) != NULL ) {

				/* Don't ascend up the tree or include the current directory */
				if ( strcmp( entry->d_name, ".." ) != 0 && strcmp( entry->d_name, "." ) != 0 ) {

					if ( entry->d_type == DT_DIR ) {

						/* Recurse into that folder */
						size += getSize( longest_name + entry->d_name );
					}

					else {

						/* Get the size of the current file */
						size += getSize( longest_name + entry->d_name );
					}
				}
			}

			closedir( d_fh );
			return size;
		}

		default: {

			/* Return 0 for all other nodes: chr, blk, lnk, symlink etc */
			return 0;
		}
	}

	/* Should never come till here */
	return 0;
};

mode_t DesQUtils::getMode( QString path ) {

	struct stat fileAtts;
	if ( stat( path.toLocal8Bit().constData(), &fileAtts ) != 0 ) {
		return -1;
	}

	return fileAtts.st_mode;
};

QStringList DesQUtils::recDirWalk( QString path ) {

	QStringList fileList;

	if ( not QFileInfo( path ).exists() )
		return QStringList();

	QDirIterator it( path, QDir::AllEntries | QDir::System | QDir::NoDotAndDotDot | QDir::Hidden, QDirIterator::Subdirectories );
	while ( it.hasNext() ) {
		it.next();
		if( it.fileInfo().isDir() ) {
			fileList.append( it.fileInfo().filePath() );
		}

		else {
			fileList.append( it.fileInfo().filePath() );
		}
	}

	return fileList;
};

QString DesQUtils::formatSize( qint64 size ) {

	if      ( size >= Size::TiB ) return QString( "%1 TiB" ).arg( QString::number( qreal( size ) / Size::TiB, 'f', 3 ) );
	else if ( size >= Size::GiB ) return QString( "%1 GiB" ).arg( QString::number( qreal( size ) / Size::GiB, 'f', 2 ) );
	else if ( size >= Size::MiB ) return QString( "%1 MiB" ).arg( QString::number( qreal( size ) / Size::MiB, 'f', 1 ) );
	else if ( size >= Size::KiB ) return QString( "%1 KiB" ).arg( QString::number( qreal( size ) / Size::KiB, 'f', 1 ) );
	else                          return QString( "%1 B"   ).arg( size );

	return QString();
};

double DesQUtils::formatSizeRaw( qint64 size ) {

	if      ( size >= Size::TiB ) return qreal( size ) / Size::TiB;
	else if ( size >= Size::GiB ) return qreal( size ) / Size::GiB;
	else if ( size >= Size::MiB ) return qreal( size ) / Size::MiB;
	else if ( size >= Size::KiB ) return qreal( size ) / Size::KiB;
	else                          return size;

	return size;
};

QString DesQUtils::formatSizeStr( qint64 size ) {

	if      ( size >= Size::TiB ) return QString( "TiB" );
	else if ( size >= Size::GiB ) return QString( "GiB" );
	else if ( size >= Size::MiB ) return QString( "MiB" );
	else if ( size >= Size::KiB ) return QString( "KiB" );
	else                          return QString( "B" );

	return "B";
};

quint64 DesQUtils::maxYForSize( quint64 fileSize ) {

	QList<quint64> sizes = { 1ULL << 0, 1ULL << 10, 1ULL << 20, 1ULL << 30, 1ULL << 40 };
    QList<quint64> widths = { 1, 2, 5, 10, 20, 50, 100, 200, 500 };

	Q_FOREACH( quint64 size, sizes ) {
        Q_FOREACH( quint64 width, widths ) {
			if ( fileSize < width * size )
				return width * size;
		}
	}

	return 0;
};

QString DesQUtils::getUtilityPath( QString utility ) {

	QFile uf( PkgConfigPath + utility + ".conf" );
	uf.open( QFile::ReadOnly );
	QString path = QString::fromUtf8( uf.readAll().simplified() );

	return path;
};
