/*
	*
	* This file is a part of DesQSession.
	* DesQSession is a Session manager for DesQ Shell
    * Copyright 2020 Marcus Britanicus
	*

	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* (at your option) any later version.
	*

	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	* GNU General Public License for more details.
	*

	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, write to the Free Software
	* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	* MA 02110-1301, USA.
	*
*/

#include <unistd.h>
#include <QStorageInfo>

#include "DesQXdg.hpp"
#include "DesQUtils.hpp"

QStringList appsDirs = QStringList() << QDir::home().filePath( ".local/share/applications/" )
							<< "/usr/local/share/applications/" << "/usr/share/applications/"
							<< "/usr/share/applications/kde4/" << "/usr/share/gnome/applications/";

DesQDesktopFile::DesQDesktopFile( QString filename ) {

	if ( filename.isEmpty() )
		return;

	mFileUrl = desktopPathForName( filename );
	mDesktopName = DesQUtils::baseName( filename );

	if ( mFileUrl.startsWith( "/usr/share/applications" ) )
		mRank = 1;

	else if ( mFileUrl.startsWith( "/usr/local/share/applications" ) )
		mRank = 2;

	else if ( mFileUrl.startsWith( QDir::home().filePath( ".local/share/applications" ) ) )
		mRank = 3;

	QSettings s( mFileUrl, QSettings::NativeFormat );
	s.beginGroup( "Desktop Entry" );

	mName = s.value( "Name" ).toString();
	mGenericName = s.value( "GenericName", mName ).toString();
	mDescription = s.value( "Comment", "No description available" ).toString();
	mCommand = s.value( "Exec" ).toString();
	mExec = s.value( "TryExec", mCommand.split( " ", Qt::SkipEmptyParts ).value( 0 ) ).toString();

	if ( not DesQUtils::exists( mExec ) ) {
		Q_FOREACH( QString path, QString::fromLocal8Bit( qgetenv( "PATH" ) ).split( ":", Qt::SkipEmptyParts ) ) {
			if ( DesQUtils::exists( path + "/" + mExec ) ) {
				mExec = path + "/" + mExec;
				break;
			}
		}
	}

	mIcon = s.value( "Icon" ).toString();

	QStringList args = mCommand.split( " " );
	foreach( QString arg, args ) {
		if ( arg == "%f" or arg == "%u" ) {
			mMultiArgs = false;
			mTakesArgs = true;
			mParsedArgs << "<#DESQARG-FILE#>";
		}

		else if ( arg == "%F" or arg == "%U" ) {
			mMultiArgs = true;
			mTakesArgs = true;
			mParsedArgs << "<#DESQARG-FILES#>";
		}

		else if ( arg == "%i" ) {
			if ( !mIcon.isEmpty() )
				mParsedArgs << "--icon" << mIcon;
		}

		else if ( arg == "%c" )
			mParsedArgs << mName;

		else if ( arg == "%k" )
			mParsedArgs << QUrl( mFileUrl ).toLocalFile();

		else
			mParsedArgs << arg;
	}

	QFile desktop( mFileUrl );
	desktop.open( QFile::ReadOnly );
	QStringList lines = QString::fromLocal8Bit( desktop.readAll() ).split( "\n" );
	Q_FOREACH( QString line, lines ) {
		if ( line.startsWith( "MimeType=" ) )
			mMimeTypes = line.replace( "MimeType=", "" ).split( ";", Qt::SkipEmptyParts );

		if ( line.startsWith( "Categories=" ) )
			mCategories = line.replace( "Categories=", "" ).split( ";", Qt::SkipEmptyParts );
	}
	desktop.close();

	getCategory();

	mVisible = not s.value( "NoDisplay", false ).toBool();
	mRunInTerminal = s.value( "Terminal", false ).toBool();

	if ( s.value( "Type" ).toString() == "Application" )
		mType = Application;

	else if ( s.value( "Type" ).toString() == "Link" )
		mType = Link;

	else if ( s.value( "Type" ).toString() == "Directory" )
		mType = Directory;

	if ( mName.count() and mCommand.count() ) {
		if ( not DesQUtils::exists( mExec ) )
			mValid = false;

		else
			mValid = true;
	}
};

bool DesQDesktopFile::startApplication() {

	if ( not mValid )
		return false;

	return QProcess::startDetached( mExec, {} );
};

bool DesQDesktopFile::startApplicationWithArgs( QStringList args ) {

	if ( not mValid )
		return false;

	QProcess proc;

	QStringList execList;

	/* If this is a commandline app */
	if ( runInTerminal() )
		execList << QFileInfo( "/etc/alternatives/x-terminal-emulator" ).symLinkTarget() << "-e";

	execList << parsedExec();
	QString exec = execList.takeFirst();

	if ( not args.count() ) {

		execList.removeAll( "<#DESQARG-FILES#>" );
		execList.removeAll( "<#DESQARG-FILE#>" );

		return QProcess::startDetached( exec, execList );
	}

	QStringList argList;
	if ( mTakesArgs ) {
		if ( mMultiArgs ) {
			Q_FOREACH( QString exeArg, execList ) {
				if ( exeArg == "<#DESQARG-FILES#>" ) {
					if ( args.count() )
						argList << args;
				}

				else
					argList << exeArg;
			}
		}

		else {
			int idx = exec.indexOf( "<#DESQARG-FILE#>" );
			argList << execList;
			argList.removeAt( idx );
			if ( args.count() ) {
				argList.insert( idx, args.takeAt( 0 ) );
				argList << args;
			}
		}
	}

	else {
		argList << execList;
		if ( args.count() )
			argList << args;
	}

	return QProcess::startDetached( exec, argList );
};

QString DesQDesktopFile::desktopName() const {

	return mDesktopName;
};

QString DesQDesktopFile::name() const {

	return mName;
};

QString DesQDesktopFile::genericName() const {

	return mGenericName;
};

QString DesQDesktopFile::description() const {

	return mDescription;
};

QString DesQDesktopFile::exec() const {

	return mExec;
};

QString DesQDesktopFile::command() const {

	return mCommand;
};

QString DesQDesktopFile::icon() const {

	return mIcon;
};

QString DesQDesktopFile::category() const {

	return mCategory;
};

QStringList DesQDesktopFile::mimeTypes() const {

	return mMimeTypes;
};

QStringList DesQDesktopFile::categories() const {

	return mCategories;
};

QStringList DesQDesktopFile::parsedExec() const {
	return mParsedArgs;
};

int DesQDesktopFile::type() const{

	return mType;
};

int DesQDesktopFile::rank() const {

	return mRank;
};

bool DesQDesktopFile::visible() const {

	return mVisible;
};

bool DesQDesktopFile::runInTerminal() const {

	return mRunInTerminal;
};

bool DesQDesktopFile::isValid() const {

	return mValid;
};

QString DesQDesktopFile::desktopFileUrl() const {

	return mFileUrl;
};

bool DesQDesktopFile::operator==( const DesQDesktopFile& other ) const {

	bool truth = true;
	truth &= ( mCommand == other.command() );
	truth &= ( mRank == other.rank() );

	return truth;
};

QString DesQDesktopFile::desktopPathForName( QString desktopName ) {

	if ( not desktopName.endsWith( ".desktop" ) )
		desktopName += ".desktop";

	if ( DesQUtils::exists( desktopName ) )
		return desktopName;

	Q_FOREACH( QString appDirStr, appsDirs ) {
		if ( DesQUtils::exists( appDirStr + "/" + desktopName ) ) {
			return appDirStr + "/" + desktopName;
		}
	}

	return QString();
};

void DesQDesktopFile::getCategory() {

	QStringList Accessories = QStringList() << "Utility" << "Utilities" << "Accessory" << "Accessories";
	QStringList Development = QStringList() << "Development";
	QStringList Education = QStringList() << "Education";
	QStringList Games = QStringList() << "Games" << "Game" << "ArcadeGame" << "StrategyGame" << "LogicGame";
	QStringList Graphics = QStringList() << "Graphics";
	QStringList Internet = QStringList() << "Network" << "Internet";
	QStringList Multimedia = QStringList() << "Audio" << "Video" << "AudioVideo" << "Multimedia";
	QStringList Office = QStringList() << "Office";
	QStringList ScienceMath = QStringList() << "Science" << "Math";
	QStringList Settings = QStringList() << "Settings";
	QStringList System = QStringList() << "System";
	QStringList Wine = QStringList() << "Wine";

	/* If the file name and the command contains 'wine', then it's a wine application */
	if ( mFileUrl.contains( "wine", Qt::CaseInsensitive ) and mCommand.contains( "wine", Qt::CaseInsensitive ) ) {
		mCategories = QStringList( { "Wine" } );
		mCategory = "Wine";
		return;
	}

	else {
		mCategory = "Uncategorized";
		foreach( QString cate, mCategories ) {
			if ( Accessories.contains( cate ) )
				mCategory = "Accessories";

			else if ( Development.contains( cate ) )
				mCategory = "Development";

			else if ( ScienceMath.contains( cate ) )
				mCategory = "Science & Math";

			else if ( Education.contains( cate ) )
				mCategory = "Education";

			else if ( Games.contains( cate ) )
				mCategory = "Games";

			else if ( Office.contains( cate ) )
				mCategory = "Office";

			else if ( Graphics.contains( cate ) )
				mCategory = "Graphics";

			else if ( Internet.contains( cate ) )
				mCategory = "Internet";

			else if ( Multimedia.contains( cate ) )
				mCategory = "Multimedia";

			else if ( Settings.contains( cate ) )
				mCategory = "Settings";

			else if ( System.contains( cate ) )
				mCategory = "System";

			else if ( Wine.contains( cate ) )
				mCategory = "Wine";

			else
				continue;
		}
	}
};

uint qHash( const DesQDesktopFile &app ) {

	QString hashString;
	hashString += app.name();
	hashString += app.genericName();
	hashString += app.description();
	hashString += app.command();
	hashString += app.icon();
	hashString += app.mimeTypes().join( " " );
	hashString += app.categories().join( " " );

	return qChecksum( hashString.toLocal8Bit().data(), hashString.count() );
};
